header {
    /* import language specific stuff */
}

class lexer extends Lexer("lexer0");

options {
    k = 2;
    charVocabulary = '\3'..'\377';
    caseSensitive=true;
    testLiterals = true;
    codeGenMakeSwitchThreshold = 2;
    codeGenBitsetTestThreshold = 3;
    exportVocab=lexer;
    importVocab=parser;
}

tokens {
    DOTDOT;
    ELLIPSIS;
}

ASSIGN_OP           :   "::="   ;
BAR                 :   '|'     ;
COLON               :   ':'     ;
COMMA               :   ','     ;
DOT                 :   '.'     ;

DOTDOT              
    :   ("...")   => "..."  { $setType(ELLIPSIS);   }
    |   ".."                { $setType(DOTDOT);     }
    ;

EXCLAMATION         :   '!'     ;
INTERSECTION        :   '^'     ;
LESS                :   '<'     ;
L_BRACE             :   '{'     ;
L_BRACKET           :   '['     ;
LL_BRACKET          :  { state_with_syntax==false }?  "[["    ;
L_PAREN             :   '('     ;
MINUS               :   '-'     ;
PLUS                :   '+'     ;
R_BRACE             :   '}'     ;
R_BRACKET           :   ']'     ;
RR_BRACKET          :  { state_with_syntax==false }?  "]]"    ;
R_PAREN             :   ')'     ;
SEMI                :   ';'     ;
AT                  :   '@'     ;


/* These are whitespace (without newline) characters according to X.680:2002 */
protected
WSchr
    : '\u0009'   // horizontal tab (HT)  '\t' 0x09  9
    | '\u0020'   // space          (SP)  ' '  0x20 32
    ;    

/* Same as WSign - just ignore consumed character */
protected
WSign  
    : WSchr { $setText(""); }
    ;

/* the end of line */
protected
EOLchr
    :   ( 
            options {
                generateAmbigWarnings = false;
            }
        : '\r''\n'
        | '\r'    
        | '\n'
        |  '\u000B'   // vertical   tab (VT)       0x0b 11
        |  '\u000C'   // form feed      (FF)  '\f' 0x0c 12
        )   
        {
            /* go ahead and bump line counter */
            newline(); 
        }
    ;

/* like EOL but we ignore the consumed symbol */
protected
EOLign
    : EOLchr { $setText("");  }
    ;

/* like EOL but we normalize consumed symbol */
protected
EOLnrm
    : EOLchr { $setText("\n");  }
    ;

/* upper (ASCII) case characters */
protected
UPCHR 
    : 'A' .. 'Z' 
    ;

/* lower (ASCII) case characters */
protected
LOCHR 
    : 'a' .. 'z' 
    ;

/* what's a (arabic) digit */
protected
DIGIT 
    : '0' .. '9' 
    ;

/* whats a (roman) letter - yes, the name sucks a bit */
protected
CHR 
    : UPCHR | LOCHR 
    ;

/* what's allowed in an identifier */
protected
IDCHR 
    : CHR | '-' | DIGIT 
    ;


/* a binary digit */
protected
BINCHR      
    :   ('0'|'1') 
    ;

/* a hex digit */
protected
HEXCHR      
    :   ('0'..'9')
    |   ('A'..'F')
    |   ('a'..'f')
    ;

/* a binary string */
protected
BINSTR  
    : "'" (BINCHR|WSign|EOLign)+ "'B" ;

/* a hex string */
protected
HEXSTR  
    : "'" (HEXCHR|WSign|EOLign)+ "'H"  ;

/* escape character in character strings */
protected
CHResc
    : '"' '"' { $setText("\""); }
    ;


/* define which input symbols we can skip (so called whitespace) */
WS          
    : ( WSchr | EOLchr )+   { $setType(Token.SKIP); }
    ;


/* A number is a sequence of digits - note that deliberatly we allow
** here for tokens like '001' etc. 
*/
TOKEN_NUMBER  
    : (DIGIT)+ 
    ;


/* what's an idenifier */
ID  
{ boolean lowchrseen=false;}
    : ("BIT" WS "STRING") => "BIT" WS "STRING"        { 
            $setType(TOKEN_BIT_STRING);   
        }
    | ("OCTET" WS "STRING") => "OCTET" WS "STRING"    { 
            $setType(TOKEN_OCTET_STRING); 
        }
    | ("OBJECT" WS "IDENTIFIER") => "OBJECT" WS "IDENTIFIER" {
            $setType(TOKEN_OBJECT_IDENTIFIER);
        }
    | ("ENCODED" WS "BY") => "ENCODED" WS "BY" {
            $setType(TOKEN_ENCODED_BY);
        }
    | ("CONSTRAINED" WS "BY") => "CONSTRAINED" WS "BY" {
            $setType(TOKEN_CONSTRAINED_BY);
        }
    | ("DEFINED" WS "BY") => "DEFINED" WS "BY" {
            $setType(TOKEN_DEFINED_BY);
        }
    | UPCHR ( LOCHR{lowchrseen=true;}|UPCHR|DIGIT|'-')* { 
            if(lowchrseen) {
                $setType(TOKEN_Word);
            } else {
                $setType(TOKEN_WORD); 
            }
        }
    |  LOCHR ( IDCHR )* { 
            $setType(TOKEN_word); 
        }
    ;   

/* what's a field */
FIELD  
{ boolean lowchrseen=false;}
    :  '&' UPCHR ( LOCHR{lowchrseen=true;}|UPCHR|DIGIT|'-')* { 
            if(lowchrseen) {
                $setType(TOKEN_Field);
            } else {
                $setType(TOKEN_FIELD); 
            }
        }
    |  '&' LOCHR ( IDCHR )* { $setType(TOKEN_field); }
    ;   



/* an octet string is either a bit string or a hex string */
OCTSTR
    : (BINSTR)=>BINSTR { $setType(TOKEN_BSTRING);}
    | HEXSTR           { $setType(TOKEN_HSTRING);}
    ;


/* A character string: this rule is not 1oo% correct as it will not
** ignore ws before  and  after eol. This needs  best to be handled
** via a language specific function. Note  that  rule  EOLnrm  will 
** replace any eol character by \n to simplify text processing.
** Contrary, ws is not normalized as ws can't be ignored in general.
*/
TOKEN_CSTRING 
    :   '"' (CHResc | EOLnrm | ~('"'|'\r'|'\n'|'\u000b'|'\u000c'))* '"' {
            String s = chr_ws_erase($getText,"\n","\t ");
            $setText(s);
        }
    ;



/* ASN.1 has kind of tricky comment rule: A comment starts with "--"
** and ends either with a "--" or with a eol character. Nesting of
** comments is therefore not possible, ie.
**  -- not visible -- visible -- not visible
** The real ugly thing about this is that you can't just uncomment
** a line (regardless of it's content) by prefixing the liene with
** "--". For example assume you have this line:
**  one INTEGER ::= 1  -- sample integer
** Then have this:
**  -- one INTEGER ::= 1  -- sample integer
** This will hide ASN.1 and just makes the comment visible!
*/

COMMENT
    :
        "--" 
        (
            ~('-'|'\n'|'\r'|'\u000b'|'\u000c') | {LA(2) != '-'}? '-'
        )* 
        {
            if ((LA(1)=='-')) {
                match("--");
            }
            $setType(Token.SKIP); 
        }
    ;

ALTCOMMENT
    : { altcomment == true }? 
        ( ALTCOMMENT1 
        | ALTCOMMENT2 
        | ALTCOMMENT3
        )
        {
            $setType(Token.SKIP); 
        }
    ;

/* Due to problematic ASN.1  commentaries we have an alternative - 
** "//" starts a comment that eat's up everything till end of line
** (as in C++ and Java).
*/

protected
ALTCOMMENT1
    : 
        { altcomment == true }? "//" (~('\n'|'\r'|'\u000b'|'\u000c'))*
        {
            // note that line counter doesn't need to be bumped here. In
            // fact doing so will lead to a wrong line number. When this
            // rule exits, we have LA(1)  pointing  to  a EOL  character.
            // But a public rule eating ws exists  already (see rule WS),
            // and this rule also handles the counting.
        }
    ;

/* We also also for typical C comments albeit not nested ones */
protected
ALTCOMMENT2
    : "/*" 
        (
            options { 
                greedy=false; 
            } 
        : '\r' ( options { warnWhenFollowAmbig=false; } : '\n')? { newline(); }
        | '\n' { newline(); }
        | .
        )* 
        "*/" 
        { }
    ;

/* And as homage to the master of style, Niklaus Wirth, we also also
** comments ala PASCAL */
protected
ALTCOMMENT3
    : "{*" 
        (
            options { 
                greedy=false; 
            } 
        : '\r' ( options { warnWhenFollowAmbig=false; } : '\n')? { newline(); }
        | '\n' { newline(); }
        | .
        )* 
        "*}" 
        {
            //$setType(Token.SKIP);  
        }
    ;

